#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
"""Commandline API gateway for laskea."""
import copy
import json
import os
import pathlib
import sys
from typing import Dict, List, Tuple, Union, no_type_check

import jmespath
import scooby  # type: ignore
import typer
from cogapp import Cog, CogUsageError  # type: ignore

import laskea
import laskea.api.jira as api
import laskea.laskea as fill

# from laskea import login, query, table

APP_NAME = 'Calculate (Finnish: laskea) some parts.'
APP_ALIAS = 'laskea'
APP_ENV = 'ASCIINATOR'
DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
QUIET = False
ENCODING = 'utf-8'
DEFAULT_MARKERS = '[[[fill ]]] [[[end]]]'
BASE_MARKERS = os.getenv(f'{APP_ENV}_MARKERS', DEFAULT_MARKERS)
FAKE_SECRET = '*' * 13
TEMPLATE_EXAMPLE = """\
{
  "table": {
    "column": {
      "fields": [
        "Key",
        "Summary",
        "Custom Field Name",
        "Custom Field Other"
      ],
      "field_map": {
        "key": [
          "key",
          "key"
        ],
        "summary": [
          "summary",
          "fields.summary"
        ],
        "custom field name": [
          "customfield_11501",
          "fields.customfield_11501"
        ],
        "custom field other": [
          "customfield_13901",
          "fields.customfield_13901[].value"
        ]
      }
    }
  },
  "remote": {
    "user": "",
    "token": "",
    "base_url": "https://remote-jira-instance.example.com/"
  },
  "local": {
    "markers": "[[[fill ]]] [[[end]]]",
    "verbose": false
  }
}
"""

app = typer.Typer(
    add_completion=False,
    context_settings={'help_option_names': ['-h', '--help']},
    no_args_is_help=True,
)


@app.callback(invoke_without_command=True)
def callback(
    version: bool = typer.Option(
        False,
        '-V',
        '--version',
        help='Display the laskea version and exit',
        is_eager=True,
    )
) -> None:
    """
    Calculate (Finnish: laskea) some parts.
    """
    if version:
        typer.echo(f'{APP_NAME} version {laskea.__version__}')
        raise typer.Exit()


@app.command('template')
def app_template() -> int:
    """
    Write a template of a well-formed JSON configuration to standard out and exit

    The strategy for looking up configurations is to start at the current working
    directory trying to read a file with the name `.laskea.json` else try to read
    same named file in the user folder (home).

    In case an explicit path is given to the config option of commands that offer
    it, only that path is considered.
    """
    sys.stdout.write(TEMPLATE_EXAMPLE)
    return sys.exit(0)


@app.command('report')
def report() -> int:
    """
    Write a report of the environment for bug reports to standard out and exit
    """

    class Report(scooby.Report):  # type: ignore
        def __init__(self, additional=None, ncol=3, text_width=80, sort=False):  # type: ignore
            """Initiate a scooby.Report instance."""

            # Mandatory packages.
            core = [
                'laskea',
                # 'antlr4',
                'atlassian',  # has version in VERSION text file in package info only?
                'cogapp.cogapp',
                'jmespath',
                'pydantic',
                'scooby',
                'typer',
            ]

            # Optional packages.
            optional: List[str] = []

            scooby.Report.__init__(
                self, additional=additional, core=core, optional=optional, ncol=ncol, text_width=text_width, sort=sort
            )

    sys.stdout.write(str(Report()))  # type: ignore
    return sys.exit(0)


@no_type_check
def load_configuration(configuration: Dict[str, object]) -> Dict[str, str]:
    """LaterAlligator."""
    if not configuration:
        print('Warning: Requested load from empty configuration')
        return {}

    source_of = {}

    column_fields = jmespath.search('table.column.fields[]', configuration)
    if column_fields:
        source_of['column_fields'] = 'config'
        api.BASE_COL_FIELDS = copy.deepcopy(column_fields)
    column_fields = os.getenv(f'{APP_ENV}_COL_FIELDS', '')
    if column_fields:
        source_of['column_fields'] = 'env'
        api.BASE_COL_FIELDS = json.loads(column_fields)

    field_map = jmespath.search('table.column.field_map', configuration)
    if field_map:
        source_of['field_map'] = 'config'
        api.BASE_COL_MAPS = copy.deepcopy(field_map)
    field_map = os.getenv(f'{APP_ENV}_COL_MAPS', '')
    if field_map:
        source_of['field_map'] = 'env'
        api.BASE_COL_MAPS = json.loads(field_map)

    remote_user = jmespath.search('remote.user', configuration)
    if remote_user:
        source_of['remote_user'] = 'config'
        api.BASE_USER = remote_user
    remote_user = os.getenv(f'{APP_ENV}_USER', '')
    if remote_user:
        source_of['remote_user'] = 'env'
        api.BASE_USER = remote_user

    remote_token = jmespath.search('remote.token', configuration)
    if remote_token:
        source_of['remote_token'] = 'config'
        api.BASE_TOKEN = remote_token
    remote_token = os.getenv(f'{APP_ENV}_TOKEN', '')
    if remote_token:
        source_of['remote_token'] = 'env'
        api.BASE_TOKEN = remote_token

    remote_base_url = jmespath.search('remote.base_url', configuration)
    if remote_base_url:
        source_of['remote_base_url'] = 'config'
        api.BASE_URL = remote_base_url
    remote_base_url = os.getenv(f'{APP_ENV}_BASE_URL', '')
    if remote_base_url:
        source_of['remote_base_url'] = 'env'
        api.BASE_URL = remote_base_url

    global BASE_MARKERS
    local_markers = jmespath.search('local.markers', configuration)
    if local_markers:
        source_of['local_markers'] = 'config'
        BASE_MARKERS = local_markers
    local_markers = os.getenv(f'{APP_ENV}_MARKERS', '')
    if local_markers:
        source_of['local_markers'] = 'env'
        BASE_MARKERS = local_markers

    global DEBUG
    verbose = bool(jmespath.search('local.verbose', configuration))
    if verbose:
        source_of['verbose'] = 'config'
        DEBUG = verbose
    verbose = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
    if verbose:
        source_of['verbose'] = 'env'
        DEBUG = verbose

    return source_of


@no_type_check
def discover_configuration(conf: str) -> Tuple[Dict[str, object], str]:
    """Try to retrieve the configuration following the "(explicit, local, parents, home)
    first wun wins" strategy."""
    configuration = None
    if conf:
        cp = pathlib.Path(conf)
        if not cp.is_file() or not cp.stat().st_size:
            print('Given configuration path is no file or empty')
            sys.exit(2)
        if not QUIET:
            print(f'Reading configuration file {cp} as requested...')
        configuration = json.load(cp.open())
    else:
        cn = fill.DEFAULT_CONFIG_NAME
        cwd = pathlib.Path.cwd().resolve()
        for pp in cwd.parents:
            cp = pp / cn
            if cp.is_file() and cp.stat().st_size:
                if not QUIET:
                    print(f'Reading from discovered configuration path {cp}')
                configuration = json.load(cp.open())
                return configuration, str(cp)

        cp = pathlib.Path.home() / fill.DEFAULT_CONFIG_NAME
        if cp.is_file() and cp.stat().st_size:
            if not QUIET:
                print(f'Reading configuration file {cp} from home directory at {pathlib.Path.home()} ...')
            configuration = json.load(cp.open())
            return configuration, str(cp)

        if not QUIET:
            print(f'User home configuration path to {cp} is no file or empty - ignoring configuration data')

    return configuration, str(cp)


@no_type_check
def report_context(command: str, transaction_mode: str, vector: List[str]) -> None:
    """DRY."""
    if QUIET:
        return
    print(f'Command: ({command})', file=sys.stderr)
    print(f'- Transaction mode: ({transaction_mode})', file=sys.stderr)
    print('Environment(variable values):', file=sys.stderr)
    app_env_user = f'{APP_ENV}_USER'
    app_env_token = f'{APP_ENV}_TOKEN'
    app_env_base_url = f'{APP_ENV}_BASE_URL'
    app_env_col_fields = f'{APP_ENV}_COL_FIELDS'
    app_env_col_maps = f'{APP_ENV}_COL_MAPS'
    app_env_markers = f'{APP_ENV}_MARKERS'
    empty = ''
    print(f'- {APP_ENV}_USER: ({os.getenv(app_env_user, empty)})', file=sys.stderr)
    print(f'- {APP_ENV}_TOKEN: ({FAKE_SECRET if len(os.getenv(app_env_token, empty)) else empty})', file=sys.stderr)
    print(f'- {APP_ENV}_BASE_URL: ({os.getenv(app_env_base_url, empty)})', file=sys.stderr)
    print(f'- {APP_ENV}_COL_FIELDS: ({os.getenv(app_env_col_fields, empty)})', file=sys.stderr)
    print(f'- {APP_ENV}_COL_MAPS: ({os.getenv(app_env_col_maps, empty)})', file=sys.stderr)
    print(f'- {APP_ENV}_MARKERS: ({os.getenv(app_env_markers, empty)})', file=sys.stderr)
    print('Effective(variable values):', file=sys.stderr)
    print(f'- RemoteUser: ({api.BASE_USER})', file=sys.stderr)
    print(f'- RemoteToken: ({"*" * len(api.BASE_PASS)})', file=sys.stderr)
    print(f'- RemoteBaseURL: ({api.BASE_URL})', file=sys.stderr)
    print(f'- ColumnFields(table): ({api.BASE_COL_FIELDS})', file=sys.stderr)
    print(f'- ColumnMaps(remote->table): ({api.BASE_COL_MAPS})', file=sys.stderr)
    print(f'- Markers(pattern): ({BASE_MARKERS})', file=sys.stderr)
    print(f'- CallVector: ({vector})', file=sys.stderr)


@no_type_check
def report_sources_of_effective_configuration(source_of: Dict[str, str], header: str) -> None:
    """DRY."""
    if QUIET:
        return
    print(header)
    print('# --- BEGIN ---')
    print(json.dumps(source_of, indent=2))
    print('# --- E N D ---')


@no_type_check
def safe_report_configuration(configuration: Dict[str, object], header: str) -> None:
    """DRY."""
    if QUIET:
        return
    print(header)
    print('# --- BEGIN ---')
    fake_configuration = copy.deepcopy(configuration)
    if jmespath.search('remote.token', fake_configuration):
        fake_configuration['remote']['token'] = FAKE_SECRET  # noqa
    print(json.dumps(fake_configuration, indent=2))
    print('# --- E N D ---')


@no_type_check
def create_and_report_effective_configuration(header: str) -> None:
    """DRY."""
    if QUIET:
        return
    effective = {
        'table': {
            'column': {
                'fields': copy.deepcopy(api.BASE_COL_FIELDS),
                'field_map': copy.deepcopy(api.BASE_COL_MAPS),
            },
        },
        'remote': {
            'user': api.BASE_USER,
            'token': '',
            'base_url': api.BASE_URL,
        },
        'local': {
            'markers': BASE_MARKERS,
            'verbose': DEBUG,
        },
    }
    safe_report_configuration(effective, header)


@app.command('update')
def update(
    source: List[str],
    inp: str = typer.Option(
        '',
        '-i',
        '--input',
        help='Path to input file',
        metavar='<sourcepath>',
    ),
    conf: str = typer.Option(
        '',
        '-c',
        '--config',
        help=f'Path to config file (default is $HOME/{fill.DEFAULT_CONFIG_NAME})',
        metavar='<configpath>',
    ),
    verify: bool = typer.Option(
        False,
        '-n',
        '--dry-run',
        help='Dry run (default is False)',
    ),
    verbose: bool = typer.Option(
        False,
        '-v',
        '--verbose',
        help='Verbose output (default is False)',
    ),
    quiet: bool = typer.Option(
        False,
        '-q',
        '--quiet',
        help='Minimal output (default is False)',
    ),
) -> int:
    """
    Fill in some parts of the input document.

    You can set some options per evironment variables:

    \b
    * ASCIINATOR_USER='remote-user'
    * ASCIINATOR_TOKEN='remote-secret'
    * ASCIINATOR_BASE_URL='https://remote-jira-instance.example.com/'
    * ASCIINATOR_COL_FIELDS: '["Key", "Summary", "Custom Field Name"]'
    * ASCIINATOR_COL_MAPS='{"key": ["key", "key"], "summary": ["summary", "fields.summary"],
      "custom field name": ["customfield_123", "fields.customfield_123"]}'
    * ASCIINATOR_MARKERS='[[[fill ]]] [[[end]]]'
    * ASCIINATOR_DEBUG='AnythingTruthy'

    The quiet option (if given) disables any conflicting verbosity setting.
    """
    command = 'update'
    transaction_mode = 'commit' if not verify else 'dry-run'
    global QUIET
    global DEBUG

    if quiet:
        QUIET = True
        DEBUG = False

    configuration, cp = discover_configuration(conf)

    if configuration is not None:
        if DEBUG or verbose:
            safe_report_configuration(configuration, f'Loaded configuration from {cp}:')

        source_of = load_configuration(configuration)

        if DEBUG or verbose:
            report_sources_of_effective_configuration(source_of, f'Configuration source after loading from {cp}:')

        if not QUIET:
            print('Configuration interface requested - Experimental!')

        create_and_report_effective_configuration(f'Effective configuration combining {cp} and environment variables:')

    paths = inp if inp else source
    if not paths:
        print('Usage: laskea update [--help] [-v] [-c config-path] [-n] [-i] source-files*md [other.md]')
        sys.exit(2)

    vector = [
        APP_ALIAS,
        '-P',
        '-c',
        f'--markers={BASE_MARKERS}',
        '-p',
        'from laskea import *',
    ]
    if not verify:
        vector.append('-r')
    if quiet:
        vector.append('--verbosity=0')

    cog = Cog()

    if DEBUG or verbose:
        report_context(command, transaction_mode, vector)

    for path in paths:
        single_vector = vector + [path]
        try:
            cog.callableMain(single_vector)
        except CogUsageError as err:
            print('CodeGen processing usage error:', file=sys.stderr)
            print(str(err))
            return sys.exit(1)

    return sys.exit(0)


@app.command('version')
def app_version() -> None:
    """
    Display the laskea version and exit.
    """
    callback(True)


# pylint: disable=expression-not-assigned
# @app.command()
def main(argv: Union[List[str], None] = None) -> int:
    """Delegate processing to functional module."""
    argv = sys.argv[1:] if argv is None else argv
    return fill.main(argv)



# Generated from jql.g4 by ANTLR 4.9.3
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3c")
        buf.write("\u00c4\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\3\2\7\2\36\n\2\f\2\16\2!\13\2\3\2\3\2\3\3\7\3&\n")
        buf.write("\3\f\3\16\3)\13\3\3\3\3\3\6\3-\n\3\r\3\16\3.\3\3\7\3\62")
        buf.write("\n\3\f\3\16\3\65\13\3\3\3\7\38\n\3\f\3\16\3;\13\3\3\4")
        buf.write("\7\4>\n\4\f\4\16\4A\13\4\3\4\3\4\3\4\7\4F\n\4\f\4\16\4")
        buf.write("I\13\4\3\4\3\4\7\4M\n\4\f\4\16\4P\13\4\7\4R\n\4\f\4\16")
        buf.write("\4U\13\4\3\4\7\4X\n\4\f\4\16\4[\13\4\3\4\5\4^\n\4\3\5")
        buf.write("\7\5a\n\5\f\5\16\5d\13\5\3\5\3\5\5\5h\n\5\3\5\3\5\7\5")
        buf.write("l\n\5\f\5\16\5o\13\5\3\5\3\5\3\5\3\5\5\5u\n\5\3\5\5\5")
        buf.write("x\n\5\3\5\7\5{\n\5\f\5\16\5~\13\5\3\6\3\6\3\6\3\6\5\6")
        buf.write("\u0084\n\6\3\6\3\6\3\6\5\6\u0089\n\6\7\6\u008b\n\6\f\6")
        buf.write("\16\6\u008e\13\6\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3")
        buf.write("\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5\7\u00a3\n\7\3")
        buf.write("\b\3\b\3\b\3\b\3\b\5\b\u00aa\n\b\3\t\3\t\3\t\3\t\7\t\u00b0")
        buf.write("\n\t\f\t\16\t\u00b3\13\t\3\t\3\t\3\n\3\n\3\13\3\13\3\f")
        buf.write("\3\f\3\r\5\r\u00be\n\r\3\r\3\r\3\16\3\16\3\16\2\2\17\2")
        buf.write("\4\6\b\n\f\16\20\22\24\26\30\32\2\7\4\2\26\26\"\"\4\2")
        buf.write("\27\27\33\33\4\2\25\33\35%\3\2&^\5\2\25\25\30\30!!\2\u00e0")
        buf.write("\2\37\3\2\2\2\4\'\3\2\2\2\6?\3\2\2\2\bb\3\2\2\2\n\177")
        buf.write("\3\2\2\2\f\u00a2\3\2\2\2\16\u00a9\3\2\2\2\20\u00ab\3\2")
        buf.write("\2\2\22\u00b6\3\2\2\2\24\u00b8\3\2\2\2\26\u00ba\3\2\2")
        buf.write("\2\30\u00bd\3\2\2\2\32\u00c1\3\2\2\2\34\36\5\4\3\2\35")
        buf.write("\34\3\2\2\2\36!\3\2\2\2\37\35\3\2\2\2\37 \3\2\2\2 \"\3")
        buf.write("\2\2\2!\37\3\2\2\2\"#\7\2\2\3#\3\3\2\2\2$&\7\7\2\2%$\3")
        buf.write("\2\2\2&)\3\2\2\2\'%\3\2\2\2\'(\3\2\2\2(*\3\2\2\2)\'\3")
        buf.write("\2\2\2*\63\5\6\4\2+-\7\7\2\2,+\3\2\2\2-.\3\2\2\2.,\3\2")
        buf.write("\2\2./\3\2\2\2/\60\3\2\2\2\60\62\5\6\4\2\61,\3\2\2\2\62")
        buf.write("\65\3\2\2\2\63\61\3\2\2\2\63\64\3\2\2\2\649\3\2\2\2\65")
        buf.write("\63\3\2\2\2\668\7\7\2\2\67\66\3\2\2\28;\3\2\2\29\67\3")
        buf.write("\2\2\29:\3\2\2\2:\5\3\2\2\2;9\3\2\2\2<>\7\t\2\2=<\3\2")
        buf.write("\2\2>A\3\2\2\2?=\3\2\2\2?@\3\2\2\2@B\3\2\2\2A?\3\2\2\2")
        buf.write("BS\5\b\5\2CG\t\2\2\2DF\7\t\2\2ED\3\2\2\2FI\3\2\2\2GE\3")
        buf.write("\2\2\2GH\3\2\2\2HJ\3\2\2\2IG\3\2\2\2JN\5\b\5\2KM\7\n\2")
        buf.write("\2LK\3\2\2\2MP\3\2\2\2NL\3\2\2\2NO\3\2\2\2OR\3\2\2\2P")
        buf.write("N\3\2\2\2QC\3\2\2\2RU\3\2\2\2SQ\3\2\2\2ST\3\2\2\2TY\3")
        buf.write("\2\2\2US\3\2\2\2VX\7\n\2\2WV\3\2\2\2X[\3\2\2\2YW\3\2\2")
        buf.write("\2YZ\3\2\2\2Z]\3\2\2\2[Y\3\2\2\2\\^\5\n\6\2]\\\3\2\2\2")
        buf.write("]^\3\2\2\2^\7\3\2\2\2_a\7\37\2\2`_\3\2\2\2ad\3\2\2\2b")
        buf.write("`\3\2\2\2bc\3\2\2\2cg\3\2\2\2db\3\2\2\2eh\5\26\f\2fh\5")
        buf.write("\16\b\2ge\3\2\2\2gf\3\2\2\2hi\3\2\2\2im\5\f\7\2jl\7\t")
        buf.write("\2\2kj\3\2\2\2lo\3\2\2\2mk\3\2\2\2mn\3\2\2\2nt\3\2\2\2")
        buf.write("om\3\2\2\2pu\5\16\b\2qu\5\20\t\2ru\7\3\2\2su\5\32\16\2")
        buf.write("tp\3\2\2\2tq\3\2\2\2tr\3\2\2\2ts\3\2\2\2uw\3\2\2\2vx\5")
        buf.write("\30\r\2wv\3\2\2\2wx\3\2\2\2x|\3\2\2\2y{\7\n\2\2zy\3\2")
        buf.write("\2\2{~\3\2\2\2|z\3\2\2\2|}\3\2\2\2}\t\3\2\2\2~|\3\2\2")
        buf.write("\2\177\u0080\7#\2\2\u0080\u0081\7\31\2\2\u0081\u0083\5")
        buf.write("\16\b\2\u0082\u0084\t\3\2\2\u0083\u0082\3\2\2\2\u0083")
        buf.write("\u0084\3\2\2\2\u0084\u008c\3\2\2\2\u0085\u0086\7\13\2")
        buf.write("\2\u0086\u0088\5\16\b\2\u0087\u0089\t\3\2\2\u0088\u0087")
        buf.write("\3\2\2\2\u0088\u0089\3\2\2\2\u0089\u008b\3\2\2\2\u008a")
        buf.write("\u0085\3\2\2\2\u008b\u008e\3\2\2\2\u008c\u008a\3\2\2\2")
        buf.write("\u008c\u008d\3\2\2\2\u008d\13\3\2\2\2\u008e\u008c\3\2")
        buf.write("\2\2\u008f\u00a3\7\f\2\2\u0090\u00a3\7\24\2\2\u0091\u00a3")
        buf.write("\7\16\2\2\u0092\u00a3\7\17\2\2\u0093\u00a3\7\21\2\2\u0094")
        buf.write("\u00a3\7\20\2\2\u0095\u00a3\7\22\2\2\u0096\u00a3\7\23")
        buf.write("\2\2\u0097\u00a3\7\35\2\2\u0098\u0099\7\37\2\2\u0099\u00a3")
        buf.write("\7\35\2\2\u009a\u00a3\7\36\2\2\u009b\u00a3\7%\2\2\u009c")
        buf.write("\u009d\7\36\2\2\u009d\u00a3\7\37\2\2\u009e\u009f\7%\2")
        buf.write("\2\u009f\u00a3\7\37\2\2\u00a0\u00a1\7\32\2\2\u00a1\u00a3")
        buf.write("\7$\2\2\u00a2\u008f\3\2\2\2\u00a2\u0090\3\2\2\2\u00a2")
        buf.write("\u0091\3\2\2\2\u00a2\u0092\3\2\2\2\u00a2\u0093\3\2\2\2")
        buf.write("\u00a2\u0094\3\2\2\2\u00a2\u0095\3\2\2\2\u00a2\u0096\3")
        buf.write("\2\2\2\u00a2\u0097\3\2\2\2\u00a2\u0098\3\2\2\2\u00a2\u009a")
        buf.write("\3\2\2\2\u00a2\u009b\3\2\2\2\u00a2\u009c\3\2\2\2\u00a2")
        buf.write("\u009e\3\2\2\2\u00a2\u00a0\3\2\2\2\u00a3\r\3\2\2\2\u00a4")
        buf.write("\u00aa\7`\2\2\u00a5\u00aa\7_\2\2\u00a6\u00aa\5\24\13\2")
        buf.write("\u00a7\u00aa\5\26\f\2\u00a8\u00aa\5\32\16\2\u00a9\u00a4")
        buf.write("\3\2\2\2\u00a9\u00a5\3\2\2\2\u00a9\u00a6\3\2\2\2\u00a9")
        buf.write("\u00a7\3\2\2\2\u00a9\u00a8\3\2\2\2\u00aa\17\3\2\2\2\u00ab")
        buf.write("\u00ac\7\t\2\2\u00ac\u00b1\5\16\b\2\u00ad\u00ae\7\13\2")
        buf.write("\2\u00ae\u00b0\5\16\b\2\u00af\u00ad\3\2\2\2\u00b0\u00b3")
        buf.write("\3\2\2\2\u00b1\u00af\3\2\2\2\u00b1\u00b2\3\2\2\2\u00b2")
        buf.write("\u00b4\3\2\2\2\u00b3\u00b1\3\2\2\2\u00b4\u00b5\7\n\2\2")
        buf.write("\u00b5\21\3\2\2\2\u00b6\u00b7\t\4\2\2\u00b7\23\3\2\2\2")
        buf.write("\u00b8\u00b9\7\34\2\2\u00b9\25\3\2\2\2\u00ba\u00bb\t\5")
        buf.write("\2\2\u00bb\27\3\2\2\2\u00bc\u00be\t\6\2\2\u00bd\u00bc")
        buf.write("\3\2\2\2\u00bd\u00be\3\2\2\2\u00be\u00bf\3\2\2\2\u00bf")
        buf.write("\u00c0\5\32\16\2\u00c0\31\3\2\2\2\u00c1\u00c2\7\4\2\2")
        buf.write("\u00c2\33\3\2\2\2\32\37\'.\639?GNSY]bgmtw|\u0083\u0088")
        buf.write("\u008c\u00a2\u00a9\u00b1\u00bd")
        return buf.getvalue()


class jqlParser(Parser):

    grammarFileName = "jql.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "' '",
        "';'",
        "'.'",
        "'('",
        "')'",
        "','",
        "'='",
        "'*'",
        "'~'",
        "'!~'",
        "'<'",
        "'<='",
        "'>'",
        "'>='",
        "'!='",
    ]

    symbolicNames = [
        "<INVALID>",
        "FUNCTION",
        "DATETIME",
        "NUMBER",
        "WHITESPACE",
        "SCOL",
        "DOT",
        "PAR_OPEN",
        "PAR_CLOSE",
        "COMMA",
        "EQ",
        "ASTERISK",
        "CONTAINS",
        "NOT_CONTAINS",
        "LT",
        "LT_EQ",
        "GT",
        "GT_EQ",
        "NOT_EQ",
        "K_AFTER",
        "K_AND",
        "K_ASC",
        "K_BEFORE",
        "K_BY",
        "K_CHANGED",
        "K_DESC",
        "K_EMPTY",
        "K_IN",
        "K_IS",
        "K_NOT",
        "K_NULL",
        "K_ON",
        "K_OR",
        "K_ORDER",
        "K_TO",
        "K_WAS",
        "F_AFFECTED_VERSION",
        "F_APPROVALS",
        "F_ASSIGNEE",
        "F_ATTACHMENTS",
        "F_CATEGORY",
        "F_COMMENT_RANGE",
        "F_COMPONENT",
        "F_CREATED",
        "F_CREATED_DATE",
        "F_CREATOR",
        "F_CUSTOM_FIELD",
        "F_CUSTOMER_REQUEST_TYPE",
        "F_DATE",
        "F_DESCRIPTION",
        "F_DUE",
        "F_DURATION",
        "F_ENVIRONMENT",
        "F_EPIC_LINK",
        "F_FILTER",
        "F_FIX_VERSION",
        "F_ISSUE",
        "F_ISSUE_KEY",
        "F_ISSUE_TYPE",
        "F_KEY",
        "F_LABEL",
        "F_LABELS",
        "F_LAST_VIEWED",
        "F_LEVEL",
        "F_NUMBER",
        "F_ORGANIZATION",
        "F_ORIGINAL_ESTIMATE",
        "F_PARENT",
        "F_PRIORITY",
        "F_PROJECT",
        "F_RANK",
        "F_REMAINING_ESTIMATE",
        "F_REPORTER",
        "F_REQUEST_CHANNEL_TYPE",
        "F_REQUEST_LAST_ACTIVITY_TIME",
        "F_RESOLUTION",
        "F_RESOLUTION_DATE",
        "F_RESOLVED",
        "F_SLA",
        "F_SPRINT",
        "F_STATUS",
        "F_SUMMARY",
        "F_TEXT",
        "F_TIME_SPENT",
        "F_TYPE",
        "F_UPDATED",
        "F_USER",
        "F_VERSION",
        "F_VOTER",
        "F_VOTES",
        "F_WATCHER",
        "F_WATCHERS",
        "F_WORK_RATIO",
        "IDENTIFIER",
        "STRING_LITERAL",
        "COMMENT_RANGE",
        "COMMENT_LINE",
        "SPACES",
    ]

    RULE_parse = 0
    RULE_jql_stmt_list = 1
    RULE_jql_stmt = 2
    RULE_expr = 3
    RULE_ordering_term = 4
    RULE_operator = 5
    RULE_literal_value = 6
    RULE_literal_list = 7
    RULE_keyword = 8
    RULE_state_name = 9
    RULE_field = 10
    RULE_compare_dates = 11
    RULE_dates = 12

    ruleNames = [
        "parse",
        "jql_stmt_list",
        "jql_stmt",
        "expr",
        "ordering_term",
        "operator",
        "literal_value",
        "literal_list",
        "keyword",
        "state_name",
        "field",
        "compare_dates",
        "dates",
    ]

    EOF = Token.EOF
    FUNCTION = 1
    DATETIME = 2
    NUMBER = 3
    WHITESPACE = 4
    SCOL = 5
    DOT = 6
    PAR_OPEN = 7
    PAR_CLOSE = 8
    COMMA = 9
    EQ = 10
    ASTERISK = 11
    CONTAINS = 12
    NOT_CONTAINS = 13
    LT = 14
    LT_EQ = 15
    GT = 16
    GT_EQ = 17
    NOT_EQ = 18
    K_AFTER = 19
    K_AND = 20
    K_ASC = 21
    K_BEFORE = 22
    K_BY = 23
    K_CHANGED = 24
    K_DESC = 25
    K_EMPTY = 26
    K_IN = 27
    K_IS = 28
    K_NOT = 29
    K_NULL = 30
    K_ON = 31
    K_OR = 32
    K_ORDER = 33
    K_TO = 34
    K_WAS = 35
    F_AFFECTED_VERSION = 36
    F_APPROVALS = 37
    F_ASSIGNEE = 38
    F_ATTACHMENTS = 39
    F_CATEGORY = 40
    F_COMMENT_RANGE = 41
    F_COMPONENT = 42
    F_CREATED = 43
    F_CREATED_DATE = 44
    F_CREATOR = 45
    F_CUSTOM_FIELD = 46
    F_CUSTOMER_REQUEST_TYPE = 47
    F_DATE = 48
    F_DESCRIPTION = 49
    F_DUE = 50
    F_DURATION = 51
    F_ENVIRONMENT = 52
    F_EPIC_LINK = 53
    F_FILTER = 54
    F_FIX_VERSION = 55
    F_ISSUE = 56
    F_ISSUE_KEY = 57
    F_ISSUE_TYPE = 58
    F_KEY = 59
    F_LABEL = 60
    F_LABELS = 61
    F_LAST_VIEWED = 62
    F_LEVEL = 63
    F_NUMBER = 64
    F_ORGANIZATION = 65
    F_ORIGINAL_ESTIMATE = 66
    F_PARENT = 67
    F_PRIORITY = 68
    F_PROJECT = 69
    F_RANK = 70
    F_REMAINING_ESTIMATE = 71
    F_REPORTER = 72
    F_REQUEST_CHANNEL_TYPE = 73
    F_REQUEST_LAST_ACTIVITY_TIME = 74
    F_RESOLUTION = 75
    F_RESOLUTION_DATE = 76
    F_RESOLVED = 77
    F_SLA = 78
    F_SPRINT = 79
    F_STATUS = 80
    F_SUMMARY = 81
    F_TEXT = 82
    F_TIME_SPENT = 83
    F_TYPE = 84
    F_UPDATED = 85
    F_USER = 86
    F_VERSION = 87
    F_VOTER = 88
    F_VOTES = 89
    F_WATCHER = 90
    F_WATCHERS = 91
    F_WORK_RATIO = 92
    IDENTIFIER = 93
    STRING_LITERAL = 94
    COMMENT_RANGE = 95
    COMMENT_LINE = 96
    SPACES = 97

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9.3")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None

    class ParseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(jqlParser.EOF, 0)

        def jql_stmt_list(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.Jql_stmt_listContext)
            else:
                return self.getTypedRuleContext(jqlParser.Jql_stmt_listContext, i)

        def getRuleIndex(self):
            return jqlParser.RULE_parse

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterParse"):
                listener.enterParse(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitParse"):
                listener.exitParse(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitParse"):
                return visitor.visitParse(self)
            else:
                return visitor.visitChildren(self)

    def parse(self):

        localctx = jqlParser.ParseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_parse)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 29
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jqlParser.DATETIME)
                        | (1 << jqlParser.SCOL)
                        | (1 << jqlParser.PAR_OPEN)
                        | (1 << jqlParser.K_EMPTY)
                        | (1 << jqlParser.K_NOT)
                        | (1 << jqlParser.F_AFFECTED_VERSION)
                        | (1 << jqlParser.F_APPROVALS)
                        | (1 << jqlParser.F_ASSIGNEE)
                        | (1 << jqlParser.F_ATTACHMENTS)
                        | (1 << jqlParser.F_CATEGORY)
                        | (1 << jqlParser.F_COMMENT_RANGE)
                        | (1 << jqlParser.F_COMPONENT)
                        | (1 << jqlParser.F_CREATED)
                        | (1 << jqlParser.F_CREATED_DATE)
                        | (1 << jqlParser.F_CREATOR)
                        | (1 << jqlParser.F_CUSTOM_FIELD)
                        | (1 << jqlParser.F_CUSTOMER_REQUEST_TYPE)
                        | (1 << jqlParser.F_DATE)
                        | (1 << jqlParser.F_DESCRIPTION)
                        | (1 << jqlParser.F_DUE)
                        | (1 << jqlParser.F_DURATION)
                        | (1 << jqlParser.F_ENVIRONMENT)
                        | (1 << jqlParser.F_EPIC_LINK)
                        | (1 << jqlParser.F_FILTER)
                        | (1 << jqlParser.F_FIX_VERSION)
                        | (1 << jqlParser.F_ISSUE)
                        | (1 << jqlParser.F_ISSUE_KEY)
                        | (1 << jqlParser.F_ISSUE_TYPE)
                        | (1 << jqlParser.F_KEY)
                        | (1 << jqlParser.F_LABEL)
                        | (1 << jqlParser.F_LABELS)
                        | (1 << jqlParser.F_LAST_VIEWED)
                        | (1 << jqlParser.F_LEVEL)
                    )
                )
                != 0
            ) or (
                (((_la - 64)) & ~0x3F) == 0
                and (
                    (1 << (_la - 64))
                    & (
                        (1 << (jqlParser.F_NUMBER - 64))
                        | (1 << (jqlParser.F_ORGANIZATION - 64))
                        | (1 << (jqlParser.F_ORIGINAL_ESTIMATE - 64))
                        | (1 << (jqlParser.F_PARENT - 64))
                        | (1 << (jqlParser.F_PRIORITY - 64))
                        | (1 << (jqlParser.F_PROJECT - 64))
                        | (1 << (jqlParser.F_RANK - 64))
                        | (1 << (jqlParser.F_REMAINING_ESTIMATE - 64))
                        | (1 << (jqlParser.F_REPORTER - 64))
                        | (1 << (jqlParser.F_REQUEST_CHANNEL_TYPE - 64))
                        | (1 << (jqlParser.F_REQUEST_LAST_ACTIVITY_TIME - 64))
                        | (1 << (jqlParser.F_RESOLUTION - 64))
                        | (1 << (jqlParser.F_RESOLUTION_DATE - 64))
                        | (1 << (jqlParser.F_RESOLVED - 64))
                        | (1 << (jqlParser.F_SLA - 64))
                        | (1 << (jqlParser.F_SPRINT - 64))
                        | (1 << (jqlParser.F_STATUS - 64))
                        | (1 << (jqlParser.F_SUMMARY - 64))
                        | (1 << (jqlParser.F_TEXT - 64))
                        | (1 << (jqlParser.F_TIME_SPENT - 64))
                        | (1 << (jqlParser.F_TYPE - 64))
                        | (1 << (jqlParser.F_UPDATED - 64))
                        | (1 << (jqlParser.F_USER - 64))
                        | (1 << (jqlParser.F_VERSION - 64))
                        | (1 << (jqlParser.F_VOTER - 64))
                        | (1 << (jqlParser.F_VOTES - 64))
                        | (1 << (jqlParser.F_WATCHER - 64))
                        | (1 << (jqlParser.F_WATCHERS - 64))
                        | (1 << (jqlParser.F_WORK_RATIO - 64))
                        | (1 << (jqlParser.IDENTIFIER - 64))
                        | (1 << (jqlParser.STRING_LITERAL - 64))
                    )
                )
                != 0
            ):
                self.state = 26
                self.jql_stmt_list()
                self.state = 31
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 32
            self.match(jqlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Jql_stmt_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def jql_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.Jql_stmtContext)
            else:
                return self.getTypedRuleContext(jqlParser.Jql_stmtContext, i)

        def SCOL(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.SCOL)
            else:
                return self.getToken(jqlParser.SCOL, i)

        def getRuleIndex(self):
            return jqlParser.RULE_jql_stmt_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJql_stmt_list"):
                listener.enterJql_stmt_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJql_stmt_list"):
                listener.exitJql_stmt_list(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJql_stmt_list"):
                return visitor.visitJql_stmt_list(self)
            else:
                return visitor.visitChildren(self)

    def jql_stmt_list(self):

        localctx = jqlParser.Jql_stmt_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_jql_stmt_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 37
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.SCOL:
                self.state = 34
                self.match(jqlParser.SCOL)
                self.state = 39
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 40
            self.jql_stmt()
            self.state = 49
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 3, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 42
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while True:
                        self.state = 41
                        self.match(jqlParser.SCOL)
                        self.state = 44
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if not (_la == jqlParser.SCOL):
                            break

                    self.state = 46
                    self.jql_stmt()
                self.state = 51
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 3, self._ctx)

            self.state = 55
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 4, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 52
                    self.match(jqlParser.SCOL)
                self.state = 57
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 4, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Jql_stmtContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.ExprContext)
            else:
                return self.getTypedRuleContext(jqlParser.ExprContext, i)

        def PAR_OPEN(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.PAR_OPEN)
            else:
                return self.getToken(jqlParser.PAR_OPEN, i)

        def PAR_CLOSE(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.PAR_CLOSE)
            else:
                return self.getToken(jqlParser.PAR_CLOSE, i)

        def ordering_term(self):
            return self.getTypedRuleContext(jqlParser.Ordering_termContext, 0)

        def K_AND(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.K_AND)
            else:
                return self.getToken(jqlParser.K_AND, i)

        def K_OR(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.K_OR)
            else:
                return self.getToken(jqlParser.K_OR, i)

        def getRuleIndex(self):
            return jqlParser.RULE_jql_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterJql_stmt"):
                listener.enterJql_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitJql_stmt"):
                listener.exitJql_stmt(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitJql_stmt"):
                return visitor.visitJql_stmt(self)
            else:
                return visitor.visitChildren(self)

    def jql_stmt(self):

        localctx = jqlParser.Jql_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_jql_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 61
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.PAR_OPEN:
                self.state = 58
                self.match(jqlParser.PAR_OPEN)
                self.state = 63
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 64
            self.expr()
            self.state = 81
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.K_AND or _la == jqlParser.K_OR:
                self.state = 65
                _la = self._input.LA(1)
                if not (_la == jqlParser.K_AND or _la == jqlParser.K_OR):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 69
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jqlParser.PAR_OPEN:
                    self.state = 66
                    self.match(jqlParser.PAR_OPEN)
                    self.state = 71
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 72
                self.expr()
                self.state = 76
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 73
                        self.match(jqlParser.PAR_CLOSE)
                    self.state = 78
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)

                self.state = 83
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 87
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.PAR_CLOSE:
                self.state = 84
                self.match(jqlParser.PAR_CLOSE)
                self.state = 89
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 91
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jqlParser.K_ORDER:
                self.state = 90
                self.ordering_term()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def operator(self):
            return self.getTypedRuleContext(jqlParser.OperatorContext, 0)

        def field(self):
            return self.getTypedRuleContext(jqlParser.FieldContext, 0)

        def literal_value(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.Literal_valueContext)
            else:
                return self.getTypedRuleContext(jqlParser.Literal_valueContext, i)

        def literal_list(self):
            return self.getTypedRuleContext(jqlParser.Literal_listContext, 0)

        def FUNCTION(self):
            return self.getToken(jqlParser.FUNCTION, 0)

        def dates(self):
            return self.getTypedRuleContext(jqlParser.DatesContext, 0)

        def K_NOT(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.K_NOT)
            else:
                return self.getToken(jqlParser.K_NOT, i)

        def PAR_OPEN(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.PAR_OPEN)
            else:
                return self.getToken(jqlParser.PAR_OPEN, i)

        def compare_dates(self):
            return self.getTypedRuleContext(jqlParser.Compare_datesContext, 0)

        def PAR_CLOSE(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.PAR_CLOSE)
            else:
                return self.getToken(jqlParser.PAR_CLOSE, i)

        def getRuleIndex(self):
            return jqlParser.RULE_expr

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpr"):
                listener.enterExpr(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpr"):
                listener.exitExpr(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitExpr"):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)

    def expr(self):

        localctx = jqlParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_expr)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 96
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.K_NOT:
                self.state = 93
                self.match(jqlParser.K_NOT)
                self.state = 98
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 101
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 12, self._ctx)
            if la_ == 1:
                self.state = 99
                self.field()
                pass

            elif la_ == 2:
                self.state = 100
                self.literal_value()
                pass

            self.state = 103
            self.operator()
            self.state = 107
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 13, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 104
                    self.match(jqlParser.PAR_OPEN)
                self.state = 109
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 13, self._ctx)

            self.state = 114
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 14, self._ctx)
            if la_ == 1:
                self.state = 110
                self.literal_value()
                pass

            elif la_ == 2:
                self.state = 111
                self.literal_list()
                pass

            elif la_ == 3:
                self.state = 112
                self.match(jqlParser.FUNCTION)
                pass

            elif la_ == 4:
                self.state = 113
                self.dates()
                pass

            self.state = 117
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 15, self._ctx)
            if la_ == 1:
                self.state = 116
                self.compare_dates()

            self.state = 122
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 16, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 119
                    self.match(jqlParser.PAR_CLOSE)
                self.state = 124
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 16, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ordering_termContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def K_ORDER(self):
            return self.getToken(jqlParser.K_ORDER, 0)

        def K_BY(self):
            return self.getToken(jqlParser.K_BY, 0)

        def literal_value(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.Literal_valueContext)
            else:
                return self.getTypedRuleContext(jqlParser.Literal_valueContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.COMMA)
            else:
                return self.getToken(jqlParser.COMMA, i)

        def K_ASC(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.K_ASC)
            else:
                return self.getToken(jqlParser.K_ASC, i)

        def K_DESC(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.K_DESC)
            else:
                return self.getToken(jqlParser.K_DESC, i)

        def getRuleIndex(self):
            return jqlParser.RULE_ordering_term

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOrdering_term"):
                listener.enterOrdering_term(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOrdering_term"):
                listener.exitOrdering_term(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOrdering_term"):
                return visitor.visitOrdering_term(self)
            else:
                return visitor.visitChildren(self)

    def ordering_term(self):

        localctx = jqlParser.Ordering_termContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_ordering_term)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(jqlParser.K_ORDER)
            self.state = 126
            self.match(jqlParser.K_BY)
            self.state = 127
            self.literal_value()
            self.state = 129
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jqlParser.K_ASC or _la == jqlParser.K_DESC:
                self.state = 128
                _la = self._input.LA(1)
                if not (_la == jqlParser.K_ASC or _la == jqlParser.K_DESC):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()

            self.state = 138
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.COMMA:
                self.state = 131
                self.match(jqlParser.COMMA)
                self.state = 132
                self.literal_value()
                self.state = 134
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jqlParser.K_ASC or _la == jqlParser.K_DESC:
                    self.state = 133
                    _la = self._input.LA(1)
                    if not (_la == jqlParser.K_ASC or _la == jqlParser.K_DESC):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                self.state = 140
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OperatorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQ(self):
            return self.getToken(jqlParser.EQ, 0)

        def NOT_EQ(self):
            return self.getToken(jqlParser.NOT_EQ, 0)

        def CONTAINS(self):
            return self.getToken(jqlParser.CONTAINS, 0)

        def NOT_CONTAINS(self):
            return self.getToken(jqlParser.NOT_CONTAINS, 0)

        def LT_EQ(self):
            return self.getToken(jqlParser.LT_EQ, 0)

        def LT(self):
            return self.getToken(jqlParser.LT, 0)

        def GT(self):
            return self.getToken(jqlParser.GT, 0)

        def GT_EQ(self):
            return self.getToken(jqlParser.GT_EQ, 0)

        def K_IN(self):
            return self.getToken(jqlParser.K_IN, 0)

        def K_NOT(self):
            return self.getToken(jqlParser.K_NOT, 0)

        def K_IS(self):
            return self.getToken(jqlParser.K_IS, 0)

        def K_WAS(self):
            return self.getToken(jqlParser.K_WAS, 0)

        def K_CHANGED(self):
            return self.getToken(jqlParser.K_CHANGED, 0)

        def K_TO(self):
            return self.getToken(jqlParser.K_TO, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_operator

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterOperator"):
                listener.enterOperator(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitOperator"):
                listener.exitOperator(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitOperator"):
                return visitor.visitOperator(self)
            else:
                return visitor.visitChildren(self)

    def operator(self):

        localctx = jqlParser.OperatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_operator)
        try:
            self.state = 160
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 20, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 141
                self.match(jqlParser.EQ)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 142
                self.match(jqlParser.NOT_EQ)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 143
                self.match(jqlParser.CONTAINS)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 144
                self.match(jqlParser.NOT_CONTAINS)
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 145
                self.match(jqlParser.LT_EQ)
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 146
                self.match(jqlParser.LT)
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 147
                self.match(jqlParser.GT)
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 148
                self.match(jqlParser.GT_EQ)
                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 149
                self.match(jqlParser.K_IN)
                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 150
                self.match(jqlParser.K_NOT)
                self.state = 151
                self.match(jqlParser.K_IN)
                pass

            elif la_ == 11:
                self.enterOuterAlt(localctx, 11)
                self.state = 152
                self.match(jqlParser.K_IS)
                pass

            elif la_ == 12:
                self.enterOuterAlt(localctx, 12)
                self.state = 153
                self.match(jqlParser.K_WAS)
                pass

            elif la_ == 13:
                self.enterOuterAlt(localctx, 13)
                self.state = 154
                self.match(jqlParser.K_IS)
                self.state = 155
                self.match(jqlParser.K_NOT)
                pass

            elif la_ == 14:
                self.enterOuterAlt(localctx, 14)
                self.state = 156
                self.match(jqlParser.K_WAS)
                self.state = 157
                self.match(jqlParser.K_NOT)
                pass

            elif la_ == 15:
                self.enterOuterAlt(localctx, 15)
                self.state = 158
                self.match(jqlParser.K_CHANGED)
                self.state = 159
                self.match(jqlParser.K_TO)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Literal_valueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING_LITERAL(self):
            return self.getToken(jqlParser.STRING_LITERAL, 0)

        def IDENTIFIER(self):
            return self.getToken(jqlParser.IDENTIFIER, 0)

        def state_name(self):
            return self.getTypedRuleContext(jqlParser.State_nameContext, 0)

        def field(self):
            return self.getTypedRuleContext(jqlParser.FieldContext, 0)

        def dates(self):
            return self.getTypedRuleContext(jqlParser.DatesContext, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_literal_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLiteral_value"):
                listener.enterLiteral_value(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLiteral_value"):
                listener.exitLiteral_value(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLiteral_value"):
                return visitor.visitLiteral_value(self)
            else:
                return visitor.visitChildren(self)

    def literal_value(self):

        localctx = jqlParser.Literal_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_literal_value)
        try:
            self.state = 167
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jqlParser.STRING_LITERAL]:
                self.enterOuterAlt(localctx, 1)
                self.state = 162
                self.match(jqlParser.STRING_LITERAL)
                pass
            elif token in [jqlParser.IDENTIFIER]:
                self.enterOuterAlt(localctx, 2)
                self.state = 163
                self.match(jqlParser.IDENTIFIER)
                pass
            elif token in [jqlParser.K_EMPTY]:
                self.enterOuterAlt(localctx, 3)
                self.state = 164
                self.state_name()
                pass
            elif token in [
                jqlParser.F_AFFECTED_VERSION,
                jqlParser.F_APPROVALS,
                jqlParser.F_ASSIGNEE,
                jqlParser.F_ATTACHMENTS,
                jqlParser.F_CATEGORY,
                jqlParser.F_COMMENT_RANGE,
                jqlParser.F_COMPONENT,
                jqlParser.F_CREATED,
                jqlParser.F_CREATED_DATE,
                jqlParser.F_CREATOR,
                jqlParser.F_CUSTOM_FIELD,
                jqlParser.F_CUSTOMER_REQUEST_TYPE,
                jqlParser.F_DATE,
                jqlParser.F_DESCRIPTION,
                jqlParser.F_DUE,
                jqlParser.F_DURATION,
                jqlParser.F_ENVIRONMENT,
                jqlParser.F_EPIC_LINK,
                jqlParser.F_FILTER,
                jqlParser.F_FIX_VERSION,
                jqlParser.F_ISSUE,
                jqlParser.F_ISSUE_KEY,
                jqlParser.F_ISSUE_TYPE,
                jqlParser.F_KEY,
                jqlParser.F_LABEL,
                jqlParser.F_LABELS,
                jqlParser.F_LAST_VIEWED,
                jqlParser.F_LEVEL,
                jqlParser.F_NUMBER,
                jqlParser.F_ORGANIZATION,
                jqlParser.F_ORIGINAL_ESTIMATE,
                jqlParser.F_PARENT,
                jqlParser.F_PRIORITY,
                jqlParser.F_PROJECT,
                jqlParser.F_RANK,
                jqlParser.F_REMAINING_ESTIMATE,
                jqlParser.F_REPORTER,
                jqlParser.F_REQUEST_CHANNEL_TYPE,
                jqlParser.F_REQUEST_LAST_ACTIVITY_TIME,
                jqlParser.F_RESOLUTION,
                jqlParser.F_RESOLUTION_DATE,
                jqlParser.F_RESOLVED,
                jqlParser.F_SLA,
                jqlParser.F_SPRINT,
                jqlParser.F_STATUS,
                jqlParser.F_SUMMARY,
                jqlParser.F_TEXT,
                jqlParser.F_TIME_SPENT,
                jqlParser.F_TYPE,
                jqlParser.F_UPDATED,
                jqlParser.F_USER,
                jqlParser.F_VERSION,
                jqlParser.F_VOTER,
                jqlParser.F_VOTES,
                jqlParser.F_WATCHER,
                jqlParser.F_WATCHERS,
                jqlParser.F_WORK_RATIO,
            ]:
                self.enterOuterAlt(localctx, 4)
                self.state = 165
                self.field()
                pass
            elif token in [jqlParser.DATETIME]:
                self.enterOuterAlt(localctx, 5)
                self.state = 166
                self.dates()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Literal_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PAR_OPEN(self):
            return self.getToken(jqlParser.PAR_OPEN, 0)

        def literal_value(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jqlParser.Literal_valueContext)
            else:
                return self.getTypedRuleContext(jqlParser.Literal_valueContext, i)

        def PAR_CLOSE(self):
            return self.getToken(jqlParser.PAR_CLOSE, 0)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jqlParser.COMMA)
            else:
                return self.getToken(jqlParser.COMMA, i)

        def getRuleIndex(self):
            return jqlParser.RULE_literal_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLiteral_list"):
                listener.enterLiteral_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLiteral_list"):
                listener.exitLiteral_list(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitLiteral_list"):
                return visitor.visitLiteral_list(self)
            else:
                return visitor.visitChildren(self)

    def literal_list(self):

        localctx = jqlParser.Literal_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_literal_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 169
            self.match(jqlParser.PAR_OPEN)
            self.state = 170
            self.literal_value()
            self.state = 175
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jqlParser.COMMA:
                self.state = 171
                self.match(jqlParser.COMMA)
                self.state = 172
                self.literal_value()
                self.state = 177
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 178
            self.match(jqlParser.PAR_CLOSE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class KeywordContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def K_AFTER(self):
            return self.getToken(jqlParser.K_AFTER, 0)

        def K_AND(self):
            return self.getToken(jqlParser.K_AND, 0)

        def K_ASC(self):
            return self.getToken(jqlParser.K_ASC, 0)

        def K_BEFORE(self):
            return self.getToken(jqlParser.K_BEFORE, 0)

        def K_BY(self):
            return self.getToken(jqlParser.K_BY, 0)

        def K_CHANGED(self):
            return self.getToken(jqlParser.K_CHANGED, 0)

        def K_DESC(self):
            return self.getToken(jqlParser.K_DESC, 0)

        def K_IN(self):
            return self.getToken(jqlParser.K_IN, 0)

        def K_IS(self):
            return self.getToken(jqlParser.K_IS, 0)

        def K_NOT(self):
            return self.getToken(jqlParser.K_NOT, 0)

        def K_NULL(self):
            return self.getToken(jqlParser.K_NULL, 0)

        def K_ON(self):
            return self.getToken(jqlParser.K_ON, 0)

        def K_OR(self):
            return self.getToken(jqlParser.K_OR, 0)

        def K_ORDER(self):
            return self.getToken(jqlParser.K_ORDER, 0)

        def K_TO(self):
            return self.getToken(jqlParser.K_TO, 0)

        def K_WAS(self):
            return self.getToken(jqlParser.K_WAS, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_keyword

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKeyword"):
                listener.enterKeyword(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKeyword"):
                listener.exitKeyword(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitKeyword"):
                return visitor.visitKeyword(self)
            else:
                return visitor.visitChildren(self)

    def keyword(self):

        localctx = jqlParser.KeywordContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_keyword)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 180
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jqlParser.K_AFTER)
                            | (1 << jqlParser.K_AND)
                            | (1 << jqlParser.K_ASC)
                            | (1 << jqlParser.K_BEFORE)
                            | (1 << jqlParser.K_BY)
                            | (1 << jqlParser.K_CHANGED)
                            | (1 << jqlParser.K_DESC)
                            | (1 << jqlParser.K_IN)
                            | (1 << jqlParser.K_IS)
                            | (1 << jqlParser.K_NOT)
                            | (1 << jqlParser.K_NULL)
                            | (1 << jqlParser.K_ON)
                            | (1 << jqlParser.K_OR)
                            | (1 << jqlParser.K_ORDER)
                            | (1 << jqlParser.K_TO)
                            | (1 << jqlParser.K_WAS)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class State_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def K_EMPTY(self):
            return self.getToken(jqlParser.K_EMPTY, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_state_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterState_name"):
                listener.enterState_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitState_name"):
                listener.exitState_name(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitState_name"):
                return visitor.visitState_name(self)
            else:
                return visitor.visitChildren(self)

    def state_name(self):

        localctx = jqlParser.State_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_state_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 182
            self.match(jqlParser.K_EMPTY)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def F_AFFECTED_VERSION(self):
            return self.getToken(jqlParser.F_AFFECTED_VERSION, 0)

        def F_APPROVALS(self):
            return self.getToken(jqlParser.F_APPROVALS, 0)

        def F_ASSIGNEE(self):
            return self.getToken(jqlParser.F_ASSIGNEE, 0)

        def F_ATTACHMENTS(self):
            return self.getToken(jqlParser.F_ATTACHMENTS, 0)

        def F_CATEGORY(self):
            return self.getToken(jqlParser.F_CATEGORY, 0)

        def F_COMMENT_RANGE(self):
            return self.getToken(jqlParser.F_COMMENT_RANGE, 0)

        def F_COMPONENT(self):
            return self.getToken(jqlParser.F_COMPONENT, 0)

        def F_CREATED(self):
            return self.getToken(jqlParser.F_CREATED, 0)

        def F_CREATED_DATE(self):
            return self.getToken(jqlParser.F_CREATED_DATE, 0)

        def F_CREATOR(self):
            return self.getToken(jqlParser.F_CREATOR, 0)

        def F_CUSTOM_FIELD(self):
            return self.getToken(jqlParser.F_CUSTOM_FIELD, 0)

        def F_CUSTOMER_REQUEST_TYPE(self):
            return self.getToken(jqlParser.F_CUSTOMER_REQUEST_TYPE, 0)

        def F_DATE(self):
            return self.getToken(jqlParser.F_DATE, 0)

        def F_DESCRIPTION(self):
            return self.getToken(jqlParser.F_DESCRIPTION, 0)

        def F_DUE(self):
            return self.getToken(jqlParser.F_DUE, 0)

        def F_DURATION(self):
            return self.getToken(jqlParser.F_DURATION, 0)

        def F_ENVIRONMENT(self):
            return self.getToken(jqlParser.F_ENVIRONMENT, 0)

        def F_EPIC_LINK(self):
            return self.getToken(jqlParser.F_EPIC_LINK, 0)

        def F_FILTER(self):
            return self.getToken(jqlParser.F_FILTER, 0)

        def F_FIX_VERSION(self):
            return self.getToken(jqlParser.F_FIX_VERSION, 0)

        def F_ISSUE(self):
            return self.getToken(jqlParser.F_ISSUE, 0)

        def F_ISSUE_KEY(self):
            return self.getToken(jqlParser.F_ISSUE_KEY, 0)

        def F_ISSUE_TYPE(self):
            return self.getToken(jqlParser.F_ISSUE_TYPE, 0)

        def F_KEY(self):
            return self.getToken(jqlParser.F_KEY, 0)

        def F_LABEL(self):
            return self.getToken(jqlParser.F_LABEL, 0)

        def F_LABELS(self):
            return self.getToken(jqlParser.F_LABELS, 0)

        def F_LAST_VIEWED(self):
            return self.getToken(jqlParser.F_LAST_VIEWED, 0)

        def F_LEVEL(self):
            return self.getToken(jqlParser.F_LEVEL, 0)

        def F_NUMBER(self):
            return self.getToken(jqlParser.F_NUMBER, 0)

        def F_ORGANIZATION(self):
            return self.getToken(jqlParser.F_ORGANIZATION, 0)

        def F_ORIGINAL_ESTIMATE(self):
            return self.getToken(jqlParser.F_ORIGINAL_ESTIMATE, 0)

        def F_PARENT(self):
            return self.getToken(jqlParser.F_PARENT, 0)

        def F_PRIORITY(self):
            return self.getToken(jqlParser.F_PRIORITY, 0)

        def F_PROJECT(self):
            return self.getToken(jqlParser.F_PROJECT, 0)

        def F_RANK(self):
            return self.getToken(jqlParser.F_RANK, 0)

        def F_REMAINING_ESTIMATE(self):
            return self.getToken(jqlParser.F_REMAINING_ESTIMATE, 0)

        def F_REPORTER(self):
            return self.getToken(jqlParser.F_REPORTER, 0)

        def F_REQUEST_CHANNEL_TYPE(self):
            return self.getToken(jqlParser.F_REQUEST_CHANNEL_TYPE, 0)

        def F_REQUEST_LAST_ACTIVITY_TIME(self):
            return self.getToken(jqlParser.F_REQUEST_LAST_ACTIVITY_TIME, 0)

        def F_RESOLUTION(self):
            return self.getToken(jqlParser.F_RESOLUTION, 0)

        def F_RESOLUTION_DATE(self):
            return self.getToken(jqlParser.F_RESOLUTION_DATE, 0)

        def F_RESOLVED(self):
            return self.getToken(jqlParser.F_RESOLVED, 0)

        def F_SLA(self):
            return self.getToken(jqlParser.F_SLA, 0)

        def F_SPRINT(self):
            return self.getToken(jqlParser.F_SPRINT, 0)

        def F_STATUS(self):
            return self.getToken(jqlParser.F_STATUS, 0)

        def F_SUMMARY(self):
            return self.getToken(jqlParser.F_SUMMARY, 0)

        def F_TEXT(self):
            return self.getToken(jqlParser.F_TEXT, 0)

        def F_TIME_SPENT(self):
            return self.getToken(jqlParser.F_TIME_SPENT, 0)

        def F_TYPE(self):
            return self.getToken(jqlParser.F_TYPE, 0)

        def F_UPDATED(self):
            return self.getToken(jqlParser.F_UPDATED, 0)

        def F_USER(self):
            return self.getToken(jqlParser.F_USER, 0)

        def F_VERSION(self):
            return self.getToken(jqlParser.F_VERSION, 0)

        def F_VOTER(self):
            return self.getToken(jqlParser.F_VOTER, 0)

        def F_VOTES(self):
            return self.getToken(jqlParser.F_VOTES, 0)

        def F_WATCHER(self):
            return self.getToken(jqlParser.F_WATCHER, 0)

        def F_WATCHERS(self):
            return self.getToken(jqlParser.F_WATCHERS, 0)

        def F_WORK_RATIO(self):
            return self.getToken(jqlParser.F_WORK_RATIO, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_field

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterField"):
                listener.enterField(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitField"):
                listener.exitField(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitField"):
                return visitor.visitField(self)
            else:
                return visitor.visitChildren(self)

    def field(self):

        localctx = jqlParser.FieldContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_field)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 184
            _la = self._input.LA(1)
            if not (
                (
                    (((_la - 36)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 36))
                        & (
                            (1 << (jqlParser.F_AFFECTED_VERSION - 36))
                            | (1 << (jqlParser.F_APPROVALS - 36))
                            | (1 << (jqlParser.F_ASSIGNEE - 36))
                            | (1 << (jqlParser.F_ATTACHMENTS - 36))
                            | (1 << (jqlParser.F_CATEGORY - 36))
                            | (1 << (jqlParser.F_COMMENT_RANGE - 36))
                            | (1 << (jqlParser.F_COMPONENT - 36))
                            | (1 << (jqlParser.F_CREATED - 36))
                            | (1 << (jqlParser.F_CREATED_DATE - 36))
                            | (1 << (jqlParser.F_CREATOR - 36))
                            | (1 << (jqlParser.F_CUSTOM_FIELD - 36))
                            | (1 << (jqlParser.F_CUSTOMER_REQUEST_TYPE - 36))
                            | (1 << (jqlParser.F_DATE - 36))
                            | (1 << (jqlParser.F_DESCRIPTION - 36))
                            | (1 << (jqlParser.F_DUE - 36))
                            | (1 << (jqlParser.F_DURATION - 36))
                            | (1 << (jqlParser.F_ENVIRONMENT - 36))
                            | (1 << (jqlParser.F_EPIC_LINK - 36))
                            | (1 << (jqlParser.F_FILTER - 36))
                            | (1 << (jqlParser.F_FIX_VERSION - 36))
                            | (1 << (jqlParser.F_ISSUE - 36))
                            | (1 << (jqlParser.F_ISSUE_KEY - 36))
                            | (1 << (jqlParser.F_ISSUE_TYPE - 36))
                            | (1 << (jqlParser.F_KEY - 36))
                            | (1 << (jqlParser.F_LABEL - 36))
                            | (1 << (jqlParser.F_LABELS - 36))
                            | (1 << (jqlParser.F_LAST_VIEWED - 36))
                            | (1 << (jqlParser.F_LEVEL - 36))
                            | (1 << (jqlParser.F_NUMBER - 36))
                            | (1 << (jqlParser.F_ORGANIZATION - 36))
                            | (1 << (jqlParser.F_ORIGINAL_ESTIMATE - 36))
                            | (1 << (jqlParser.F_PARENT - 36))
                            | (1 << (jqlParser.F_PRIORITY - 36))
                            | (1 << (jqlParser.F_PROJECT - 36))
                            | (1 << (jqlParser.F_RANK - 36))
                            | (1 << (jqlParser.F_REMAINING_ESTIMATE - 36))
                            | (1 << (jqlParser.F_REPORTER - 36))
                            | (1 << (jqlParser.F_REQUEST_CHANNEL_TYPE - 36))
                            | (1 << (jqlParser.F_REQUEST_LAST_ACTIVITY_TIME - 36))
                            | (1 << (jqlParser.F_RESOLUTION - 36))
                            | (1 << (jqlParser.F_RESOLUTION_DATE - 36))
                            | (1 << (jqlParser.F_RESOLVED - 36))
                            | (1 << (jqlParser.F_SLA - 36))
                            | (1 << (jqlParser.F_SPRINT - 36))
                            | (1 << (jqlParser.F_STATUS - 36))
                            | (1 << (jqlParser.F_SUMMARY - 36))
                            | (1 << (jqlParser.F_TEXT - 36))
                            | (1 << (jqlParser.F_TIME_SPENT - 36))
                            | (1 << (jqlParser.F_TYPE - 36))
                            | (1 << (jqlParser.F_UPDATED - 36))
                            | (1 << (jqlParser.F_USER - 36))
                            | (1 << (jqlParser.F_VERSION - 36))
                            | (1 << (jqlParser.F_VOTER - 36))
                            | (1 << (jqlParser.F_VOTES - 36))
                            | (1 << (jqlParser.F_WATCHER - 36))
                            | (1 << (jqlParser.F_WATCHERS - 36))
                            | (1 << (jqlParser.F_WORK_RATIO - 36))
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Compare_datesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dates(self):
            return self.getTypedRuleContext(jqlParser.DatesContext, 0)

        def K_ON(self):
            return self.getToken(jqlParser.K_ON, 0)

        def K_AFTER(self):
            return self.getToken(jqlParser.K_AFTER, 0)

        def K_BEFORE(self):
            return self.getToken(jqlParser.K_BEFORE, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_compare_dates

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCompare_dates"):
                listener.enterCompare_dates(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCompare_dates"):
                listener.exitCompare_dates(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCompare_dates"):
                return visitor.visitCompare_dates(self)
            else:
                return visitor.visitChildren(self)

    def compare_dates(self):

        localctx = jqlParser.Compare_datesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_compare_dates)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 187
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if ((_la) & ~0x3F) == 0 and (
                (1 << _la) & ((1 << jqlParser.K_AFTER) | (1 << jqlParser.K_BEFORE) | (1 << jqlParser.K_ON))
            ) != 0:
                self.state = 186
                _la = self._input.LA(1)
                if not (
                    (
                        ((_la) & ~0x3F) == 0
                        and (
                            (1 << _la) & ((1 << jqlParser.K_AFTER) | (1 << jqlParser.K_BEFORE) | (1 << jqlParser.K_ON))
                        )
                        != 0
                    )
                ):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()

            self.state = 189
            self.dates()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DatesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DATETIME(self):
            return self.getToken(jqlParser.DATETIME, 0)

        def getRuleIndex(self):
            return jqlParser.RULE_dates

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDates"):
                listener.enterDates(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDates"):
                listener.exitDates(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitDates"):
                return visitor.visitDates(self)
            else:
                return visitor.visitChildren(self)

    def dates(self):

        localctx = jqlParser.DatesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_dates)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 191
            self.match(jqlParser.DATETIME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long,missing-docstring,reimported,unused-import,unused-variable
import laskea.laskea as fill


def test_fill_init():
    assert fill.init() is None


def test_verify_request_wrong_number_of_args():
    assert fill.verify_request([]) == (2, 'received wrong number of arguments', [''])


def test_verify_request_bogus_command():
    assert fill.verify_request(['foo', 'bar', 'baz']) == (2, 'received unknown command', [''])


def test_verify_request_bogus_inp():
    assert fill.verify_request(['update', 'not-present', 'baz']) == (1, 'source is no file', [''])


def test_verify_request_no_config():
    assert fill.verify_request(['update', 'README.md', '']) == (2, 'configuration missing', [''])


def test_verify_request_config_as_folder():
    assert fill.verify_request(['update', 'README.md', 'laskea']) == (1, 'config (laskea) is no file', [''])


def test_verify_request_config_no_json_extension():
    assert fill.verify_request(['update', 'README.md', 'README.md']) == (1, 'config has no .json extension', [''])



# Generated from jql.g4 by ANTLR 4.9.3
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .jqlParser import jqlParser
else:
    from jqlParser import jqlParser

# This class defines a complete generic visitor for a parse tree produced by jqlParser.


class jqlVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by jqlParser#parse.
    def visitParse(self, ctx: jqlParser.ParseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#jql_stmt_list.
    def visitJql_stmt_list(self, ctx: jqlParser.Jql_stmt_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#jql_stmt.
    def visitJql_stmt(self, ctx: jqlParser.Jql_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#expr.
    def visitExpr(self, ctx: jqlParser.ExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#ordering_term.
    def visitOrdering_term(self, ctx: jqlParser.Ordering_termContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#operator.
    def visitOperator(self, ctx: jqlParser.OperatorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#literal_value.
    def visitLiteral_value(self, ctx: jqlParser.Literal_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#literal_list.
    def visitLiteral_list(self, ctx: jqlParser.Literal_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#keyword.
    def visitKeyword(self, ctx: jqlParser.KeywordContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#state_name.
    def visitState_name(self, ctx: jqlParser.State_nameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#field.
    def visitField(self, ctx: jqlParser.FieldContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#compare_dates.
    def visitCompare_dates(self, ctx: jqlParser.Compare_datesContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by jqlParser#dates.
    def visitDates(self, ctx: jqlParser.DatesContext):
        return self.visitChildren(ctx)


del jqlParser



# -*- coding: utf-8 -*-
# pylint: disable=expression-not-assigned,line-too-long
"""Calculate (Finnish: laskea) some parts."""
import json
import os
import pathlib
import sys
from typing import Dict, List, Optional, Tuple, Union, no_type_check

from atlassian import Jira  # type: ignore # noqa

import laskea.api.jira as api

DEBUG_VAR = 'LASKEA_DEBUG'
DEBUG = os.getenv(DEBUG_VAR)

ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'

DEFAULT_CONFIG_NAME = '.laskea.json'

DB: Dict[str, Union[None, Jira]] = {'handle': None}


@no_type_check
def table(query_text: str = '') -> None:
    """Public document interface."""
    if not DB.get('handle', None):
        DB['handle'] = api.login()

    print(api.markdown_table(DB['handle'], query_text))


@no_type_check
def ol(query_text: str = '') -> None:
    """Public document interface for ordered list."""
    if not DB.get('handle', None):
        DB['handle'] = api.login()

    print(api.markdown_list(DB['handle'], query_text, list_type='ol'))


@no_type_check
def ul(query_text: str = '') -> None:
    """Public document interface for unordered list."""
    if not DB.get('handle', None):
        DB['handle'] = api.login()

    print(api.markdown_list(DB['handle'], query_text))


def init() -> None:
    """Minimize boilerplate in the target documents."""
    pass


def verify_request(argv: Optional[List[str]]) -> Tuple[int, str, List[str]]:
    """Fail with grace."""
    if not argv or len(argv) != 3:
        return 2, 'received wrong number of arguments', ['']

    command, inp, config = argv

    if command not in ('update',):
        return 2, 'received unknown command', ['']

    if inp:
        if not pathlib.Path(str(inp)).is_file():
            return 1, 'source is no file', ['']

    if not config:
        return 2, 'configuration missing', ['']

    config_path = pathlib.Path(str(config))
    if not config_path.is_file():
        return 1, f'config ({config_path}) is no file', ['']
    if not ''.join(config_path.suffixes).lower().endswith('.json'):
        return 1, 'config has no .json extension', ['']

    return 0, '', argv


def main(argv: Union[List[str], None] = None) -> int:
    """Drive the lookup."""
    error, message, strings = verify_request(argv)
    if error:
        print(message, file=sys.stderr)
        return error

    command, inp, config = strings

    with open(config, 'rt', encoding=ENCODING) as handle:
        configuration = json.load(handle)

    print(f'using configuration ({configuration})')
    print(f'Would act on {command =}, {inp =}, and {config =}')
    print('NotImplemented (yet)')
    return 0



# -*- coding: utf-8 -*-
# pylint: disable=expression-not-assigned,line-too-long,missing-module-docstring
import sys

from laskea.cli import app

if __name__ == '__main__':
    sys.exit(app(prog_name='laskea'))  # pragma: no cover



# Generated from jql.g4 by ANTLR 4.9.3
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .jqlParser import jqlParser
else:
    from jqlParser import jqlParser

# This class defines a complete listener for a parse tree produced by jqlParser.
class jqlListener(ParseTreeListener):

    # Enter a parse tree produced by jqlParser#parse.
    def enterParse(self, ctx: jqlParser.ParseContext):
        pass

    # Exit a parse tree produced by jqlParser#parse.
    def exitParse(self, ctx: jqlParser.ParseContext):
        pass

    # Enter a parse tree produced by jqlParser#jql_stmt_list.
    def enterJql_stmt_list(self, ctx: jqlParser.Jql_stmt_listContext):
        pass

    # Exit a parse tree produced by jqlParser#jql_stmt_list.
    def exitJql_stmt_list(self, ctx: jqlParser.Jql_stmt_listContext):
        pass

    # Enter a parse tree produced by jqlParser#jql_stmt.
    def enterJql_stmt(self, ctx: jqlParser.Jql_stmtContext):
        pass

    # Exit a parse tree produced by jqlParser#jql_stmt.
    def exitJql_stmt(self, ctx: jqlParser.Jql_stmtContext):
        pass

    # Enter a parse tree produced by jqlParser#expr.
    def enterExpr(self, ctx: jqlParser.ExprContext):
        pass

    # Exit a parse tree produced by jqlParser#expr.
    def exitExpr(self, ctx: jqlParser.ExprContext):
        pass

    # Enter a parse tree produced by jqlParser#ordering_term.
    def enterOrdering_term(self, ctx: jqlParser.Ordering_termContext):
        pass

    # Exit a parse tree produced by jqlParser#ordering_term.
    def exitOrdering_term(self, ctx: jqlParser.Ordering_termContext):
        pass

    # Enter a parse tree produced by jqlParser#operator.
    def enterOperator(self, ctx: jqlParser.OperatorContext):
        pass

    # Exit a parse tree produced by jqlParser#operator.
    def exitOperator(self, ctx: jqlParser.OperatorContext):
        pass

    # Enter a parse tree produced by jqlParser#literal_value.
    def enterLiteral_value(self, ctx: jqlParser.Literal_valueContext):
        pass

    # Exit a parse tree produced by jqlParser#literal_value.
    def exitLiteral_value(self, ctx: jqlParser.Literal_valueContext):
        pass

    # Enter a parse tree produced by jqlParser#literal_list.
    def enterLiteral_list(self, ctx: jqlParser.Literal_listContext):
        pass

    # Exit a parse tree produced by jqlParser#literal_list.
    def exitLiteral_list(self, ctx: jqlParser.Literal_listContext):
        pass

    # Enter a parse tree produced by jqlParser#keyword.
    def enterKeyword(self, ctx: jqlParser.KeywordContext):
        pass

    # Exit a parse tree produced by jqlParser#keyword.
    def exitKeyword(self, ctx: jqlParser.KeywordContext):
        pass

    # Enter a parse tree produced by jqlParser#state_name.
    def enterState_name(self, ctx: jqlParser.State_nameContext):
        pass

    # Exit a parse tree produced by jqlParser#state_name.
    def exitState_name(self, ctx: jqlParser.State_nameContext):
        pass

    # Enter a parse tree produced by jqlParser#field.
    def enterField(self, ctx: jqlParser.FieldContext):
        pass

    # Exit a parse tree produced by jqlParser#field.
    def exitField(self, ctx: jqlParser.FieldContext):
        pass

    # Enter a parse tree produced by jqlParser#compare_dates.
    def enterCompare_dates(self, ctx: jqlParser.Compare_datesContext):
        pass

    # Exit a parse tree produced by jqlParser#compare_dates.
    def exitCompare_dates(self, ctx: jqlParser.Compare_datesContext):
        pass

    # Enter a parse tree produced by jqlParser#dates.
    def enterDates(self, ctx: jqlParser.DatesContext):
        pass

    # Exit a parse tree produced by jqlParser#dates.
    def exitDates(self, ctx: jqlParser.DatesContext):
        pass


del jqlParser



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long,missing-docstring,reimported,unused-import,unused-variable
import json

import pytest

import laskea
import laskea.api.jqlLexer  # noqa
import laskea.api.jqlListener  # noqa
import laskea.api.jqlParser  # noqa
import laskea.api.jqlVisitor  # noqa
import laskea.cli as cli


def test_foo():
    assert isinstance(laskea.api.jqlLexer.jqlLexer(), laskea.api.jqlLexer.jqlLexer)


def test_bar():
    assert isinstance(laskea.api.jqlListener.jqlListener(), laskea.api.jqlListener.jqlListener)


def test_baz():
    token_stream = laskea.api.jqlParser.TokenStream()
    assert isinstance(laskea.api.jqlParser.jqlParser(token_stream), laskea.api.jqlParser.jqlParser)


def test_quux():
    assert isinstance(laskea.api.jqlVisitor.jqlVisitor(), laskea.api.jqlVisitor.jqlVisitor)


def test_report_command(capsys):
    with pytest.raises(SystemExit):
        cli.report()
    out, err = capsys.readouterr()
    assert f'laskea : {laskea.__version__}' in out
    assert not err


def test_report_context(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = False
    assert cli.report_context(command='-', transaction_mode='+', vector=['42']) is None
    out, err = capsys.readouterr()
    assert not out
    lines = err.strip().split('\n')
    assert len(lines) == 17
    assert lines[:3] == ['Command: (-)', '- Transaction mode: (+)', 'Environment(variable values):']
    assert lines[9] == 'Effective(variable values):'
    for line in lines[3:9]:
        assert line.startswith(f'- {cli.APP_ENV}_')
    assert lines[-1] == "- CallVector: (['42'])"
    cli.QUIET = quiet_flag_restore


def test_report_context_quiet(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = True
    assert cli.report_context(command='-', transaction_mode='+', vector=['42']) is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    cli.QUIET = quiet_flag_restore


def test_report_sources_of_effective_configuration(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = False
    assert cli.report_sources_of_effective_configuration(source_of={}, header='42') is None
    out, err = capsys.readouterr()
    assert not err
    lines = out.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---', '{}', '# --- E N D ---']
    cli.QUIET = quiet_flag_restore


def test_report_sources_of_effective_configuration_quiet(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = True
    assert cli.report_sources_of_effective_configuration(source_of={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    cli.QUIET = quiet_flag_restore


def test_safe_report_configuration(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = False
    assert cli.safe_report_configuration(configuration={}, header='42') is None
    out, err = capsys.readouterr()
    assert not err
    lines = out.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---', '{}', '# --- E N D ---']
    cli.QUIET = quiet_flag_restore


def test_safe_report_configuration_quiet(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = True
    assert cli.safe_report_configuration(configuration={}, header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    cli.QUIET = quiet_flag_restore


def test_safe_report_configuration_no_leak(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = False
    thing = {'remote': {'token': 'leak'}}
    gnith = {'remote': {'token': '*************'}}
    assert cli.safe_report_configuration(configuration=thing, header='42') is None
    out, err = capsys.readouterr()
    assert not err
    lines = out.strip().split('\n')
    assert lines == ['42', '# --- BEGIN ---'] + json.dumps(gnith, indent=2).split('\n') + ['# --- E N D ---']
    cli.QUIET = quiet_flag_restore


def test_create_and_report_effective_configuration(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = False
    assert cli.create_and_report_effective_configuration(header='42') is None
    out, err = capsys.readouterr()
    assert not err
    lines = out.strip().split('\n')
    assert len(lines) == 55
    assert lines[:2] == ['42', '# --- BEGIN ---']
    assert lines[-1] == '# --- E N D ---'
    cli.QUIET = quiet_flag_restore


def test_create_and_report_effective_configuration_quiet(capsys):
    quiet_flag_restore = cli.QUIET
    cli.QUIET = True
    assert cli.create_and_report_effective_configuration(header='42') is None
    out, err = capsys.readouterr()
    assert not out
    assert not err
    cli.QUIET = quiet_flag_restore


def test_load_configuration_empty(capsys):
    quiet_flag_restore = cli.DEBUG
    base_markers_restore = cli.BASE_MARKERS
    cli.DEBUG = False
    assert cli.load_configuration(configuration={}) == {}
    out, err = capsys.readouterr()
    assert not err
    assert out.strip() == 'Warning: Requested load from empty configuration'
    cli.DEBUG = quiet_flag_restore
    cli.BASE_MARKERS = base_markers_restore


def test_load_configuration_remote_token(capsys):
    quiet_flag_restore = cli.DEBUG
    base_markers_restore = cli.BASE_MARKERS
    cli.DEBUG = False
    thing = {'remote': {'token': 'leak'}}
    gnith = {'remote_token': 'config'}
    assert cli.load_configuration(configuration=thing) == gnith
    out, err = capsys.readouterr()
    assert not out
    assert not err
    cli.DEBUG = quiet_flag_restore
    cli.BASE_MARKERS = base_markers_restore



"""Calculate (Finnish: laskea) some parts."""
from laskea.api.jira import login, markdown_list, markdown_table, query
from laskea.laskea import ol, table, ul

# [[[fill git_describe()]]]
__version__ = '2022.3.4+parent.d7131397'
# [[[end]]] (checksum: 0b17d075b5b472ebae045546aef318da)
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)
__all__ = ['login', 'markdown_table', 'markdown_list', 'query', 'ol', 'table', 'ul']



# -*- coding: utf-8 -*-
# pylint: disable=expression-not-assigned,line-too-long
"""JIRA proxy connector API for code generation."""
import copy
import json
import os
from typing import no_type_check

import jmespath
from atlassian import Jira  # type: ignore # noqa

API_BASE_URL = 'https://example.com'
APP_NAME = 'ASCIINATOR'
APP_ENV = APP_NAME

DEFAULT_COLUMN_FIELDS = ['Key', 'Summary', ['Priority', 'P'], 'Status', 'Custom Field Wun', 'Custom Field Other (CFO)']

WUN_ID = 'customfield_11501'
ANOTHER_ID = 'customfield_13901'
KNOWN_CI_FIELDS = {
    'key': ['key', 'key'],
    'summary': ['summary', 'fields.summary'],
    'priority': ['priority', 'fields.priority.name'],
    'status': ['status', 'fields.status.name'],
    'custom field name': [WUN_ID, f'fields.{WUN_ID}'],
    'custom field other': [ANOTHER_ID, f'fields.{ANOTHER_ID}[].value'],
}

BASE_USER = os.getenv(f'{APP_ENV}_USER', '')
BASE_PASS = os.getenv(f'{APP_ENV}_TOKEN', '')
BASE_URL = os.getenv(f'{APP_ENV}_BASE_URL', '')
BASE_COL_FIELDS = json.loads(os.getenv(f'{APP_ENV}_COL_FIELDS', json.dumps(DEFAULT_COLUMN_FIELDS)))
BASE_COL_MAPS = json.loads(os.getenv(f'{APP_ENV}_COL_MAPS', json.dumps(KNOWN_CI_FIELDS)))


def mock(number: int) -> int:
    """Intermediate for starting the dev env in a valid state."""
    return number


def login(user: str = '', token: str = '', url: str = '') -> Jira:
    """LatAli"""
    if not user:
        user = BASE_USER
    if not token:
        token = BASE_PASS
    if not url:
        url = BASE_URL
    if not user or not token or not url:
        raise ValueError('User, Token, and URL are all required for login.')

    return Jira(url=url, username=user, password=token)


@no_type_check
def query(handle: Jira, jql_text: str, column_fields=None) -> dict:
    """EggLayingWoolMilkDear."""

    if not column_fields:
        column_fields = BASE_COL_FIELDS

    if not jql_text.strip():
        return {
            'jql_text': jql_text,
            'error': 'Empty JIRA Query Language text detected',
        }

    completed_column_fields = []
    for entry in column_fields:
        if isinstance(entry, str):
            candidate, concept, label = entry.lower(), entry, entry
        else:
            try:
                concept, label = entry
                candidate = concept.lower()
            except TypeError:
                return {
                    'jql_text': jql_text,
                    'column_fields': column_fields,
                    'parsed_columns': completed_column_fields,
                    'error': f'The column ({entry}) is neither a string nor a pair of (concept, label)',
                }

        for field in BASE_COL_MAPS.keys():
            if field in candidate:
                completed_column_fields.append(
                    {
                        'path': BASE_COL_MAPS[field][1],
                        'id': BASE_COL_MAPS[field][0],
                        'concept': concept,
                        'label': label,
                        'field': field,
                    }
                )

    if not completed_column_fields:
        return {
            'jql_text': jql_text,
            'column_fields': column_fields,
            'error': 'Completed column fields empty (no known fields?)',
        }

    # dynamics below
    try:
        handle.user(handle.username)['name'] == handle.username
    except RuntimeError as err:
        return {
            'jql_text': jql_text,
            'column_fields': column_fields,
            'parsed_columns': completed_column_fields,
            'error': str(err),
        }

    issues = handle.jql(jql_text)
    pairs = [(col['label'], col['path']) for col in completed_column_fields]
    rows = [{label: jmespath.search(path, issue) or [''] for label, path in pairs} for issue in issues['issues']]
    return {
        'jql_text': jql_text,
        'column_fields': column_fields,
        'parsed_columns': completed_column_fields,
        'error': None,
        'rows': rows,
    }


@no_type_check
def markdown_table(handle: Jira, jql_text: str, column_fields=None) -> str:
    """Yes we can ... document later."""
    data = query(handle, jql_text, column_fields)
    if data.get('error', ''):
        return json.dumps(data, indent=2)

    if not data['rows']:
        return 'WARNING: Empty table!'

    table = copy.deepcopy(data['rows'])
    columns = list(table[0].keys())
    col_wid = {key: len(key) for key in columns}
    for slot, record in enumerate(table):
        for key, cell in record.items():
            if key.lower() == 'key':
                table[slot][key] = f'[{cell}]({BASE_URL.strip("/")}/browse/{cell})'
            if not isinstance(cell, str):
                table[slot][key] = '<br>'.join(cell)
            col_wid[key] = max(len(table[slot][key]), col_wid[key])

    header_cells = [key.ljust(col_wid[key]) for key in columns]
    header = f'| {" | ".join(header_cells)} |'

    separator_cells = ['-' * (col_wid[key] + 1) for key in columns]
    separator = f'|:{"|:".join(separator_cells)}|'

    rows = [f'| {" | ".join(str(v).ljust(col_wid[k]) for k, v in line.items())} |' for line in table]
    issues = len(table)
    summary = f'\n\n{issues} issue{"" if issues == 1 else "s"}'
    return '\n'.join([header] + [separator] + rows) + summary


@no_type_check
def markdown_list(handle: Jira, jql_text: str, column_fields=None, list_type: str = 'ul') -> str:
    """Yes we can ... document later."""
    data = query(handle, jql_text, column_fields)
    if data.get('error', ''):
        return json.dumps(data, indent=2)

    if not data['rows']:
        return ''

    items = []
    for slot, record in enumerate(data['rows']):
        k, v = '', ''
        for key, cell in record.items():
            if key.lower() not in ('key', 'summary'):
                continue
            if key.lower() == 'key':
                k = f'[{cell}]({BASE_URL.strip("/")}/browse/{cell})'
            else:
                v = cell
        items.append((k, v))

    lt = '-' if list_type == 'ul' else '1.'  # implicit 'ol'
    xl = tuple(f'{lt} {" - ".join(item)}' for item in items)
    return '\n'.join(xl) + '\n'
